# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: lidar.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'lidar.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0blidar.proto\"\xd3\x05\n\x15SickScanPointCloudMsg\x12\x35\n\x06header\x18\x01 \x01(\x0b\x32%.SickScanPointCloudMsg.SickScanHeader\x12\x0e\n\x06height\x18\x02 \x01(\r\x12\r\n\x05width\x18\x03 \x01(\r\x12>\n\x06\x66ields\x18\x04 \x01(\x0b\x32..SickScanPointCloudMsg.SickScanPointFieldArray\x12\x14\n\x0cis_bigendian\x18\x05 \x01(\r\x12\x12\n\npoint_step\x18\x06 \x01(\r\x12\x10\n\x08row_step\x18\x07 \x01(\r\x12\x37\n\x04\x64\x61ta\x18\x08 \x01(\x0b\x32).SickScanPointCloudMsg.SickScanUint8Array\x12\x10\n\x08is_dense\x18\t \x01(\r\x12\x11\n\tnum_echos\x18\n \x01(\x05\x12\x13\n\x0bsegment_idx\x18\x0b \x01(\x05\x1a^\n\x0eSickScanHeader\x12\x0b\n\x03seq\x18\x01 \x01(\r\x12\x15\n\rtimestamp_sec\x18\x02 \x01(\r\x12\x16\n\x0etimestamp_nsec\x18\x03 \x01(\r\x12\x10\n\x08\x66rame_id\x18\x04 \x01(\x0c\x1aV\n\x15SickScanPointFieldMsg\x12\x0c\n\x04name\x18\x01 \x01(\x0c\x12\x0e\n\x06offset\x18\x02 \x01(\r\x12\x10\n\x08\x64\x61tatype\x18\x03 \x01(\r\x12\r\n\x05\x63ount\x18\x04 \x01(\r\x1a\x44\n\x12SickScanUint8Array\x12\x10\n\x08\x63\x61pacity\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\x04\x12\x0e\n\x06\x62uffer\x18\x03 \x01(\x0c\x1aw\n\x17SickScanPointFieldArray\x12\x10\n\x08\x63\x61pacity\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\x04\x12<\n\x06\x62uffer\x18\x03 \x03(\x0b\x32,.SickScanPointCloudMsg.SickScanPointFieldMsgb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'lidar_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SICKSCANPOINTCLOUDMSG']._serialized_start=16
  _globals['_SICKSCANPOINTCLOUDMSG']._serialized_end=739
  _globals['_SICKSCANPOINTCLOUDMSG_SICKSCANHEADER']._serialized_start=366
  _globals['_SICKSCANPOINTCLOUDMSG_SICKSCANHEADER']._serialized_end=460
  _globals['_SICKSCANPOINTCLOUDMSG_SICKSCANPOINTFIELDMSG']._serialized_start=462
  _globals['_SICKSCANPOINTCLOUDMSG_SICKSCANPOINTFIELDMSG']._serialized_end=548
  _globals['_SICKSCANPOINTCLOUDMSG_SICKSCANUINT8ARRAY']._serialized_start=550
  _globals['_SICKSCANPOINTCLOUDMSG_SICKSCANUINT8ARRAY']._serialized_end=618
  _globals['_SICKSCANPOINTCLOUDMSG_SICKSCANPOINTFIELDARRAY']._serialized_start=620
  _globals['_SICKSCANPOINTCLOUDMSG_SICKSCANPOINTFIELDARRAY']._serialized_end=739
# @@protoc_insertion_point(module_scope)



from phenomate_core.preprocessing.base import BasePreprocessor
from phenomate_core.preprocessing.hyperspec.process import HyperspecPreprocessor
from phenomate_core.preprocessing.jai.process import JaiPreprocessor
from phenomate_core.preprocessing.oak_d.process import (
    OakCalibrationPreprocessor,
    OakFramePreprocessor,
    OakImuPacketsPreprocessor,
)

__all__ = (
    "BasePreprocessor",
    "HyperspecPreprocessor",
    "JaiPreprocessor",
    "OakCalibrationPreprocessor",
    "OakFramePreprocessor",
    "OakImuPacketsPreprocessor",
)


def get_preprocessor(sensor: str, details: str = "") -> type[BasePreprocessor]:
    match sensor.lower():
        case "jai":
            return JaiPreprocessor
        case "hyperspec" | "hyper-spec" | "hyperspectral" | "hyper-spectral":
            return HyperspecPreprocessor
        case "oak":
            if "calibration" in details:
                return OakCalibrationPreprocessor
            if "imu" in details:
                return OakImuPacketsPreprocessor
            return OakFramePreprocessor
    raise ValueError(f"Unsupported sensor type: {sensor}")



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class SickScanPointCloudMsg(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class SickScanHeader(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        SEQ_FIELD_NUMBER: builtins.int
        TIMESTAMP_SEC_FIELD_NUMBER: builtins.int
        TIMESTAMP_NSEC_FIELD_NUMBER: builtins.int
        FRAME_ID_FIELD_NUMBER: builtins.int
        seq: builtins.int
        timestamp_sec: builtins.int
        timestamp_nsec: builtins.int
        frame_id: builtins.bytes
        def __init__(
            self,
            *,
            seq: builtins.int = ...,
            timestamp_sec: builtins.int = ...,
            timestamp_nsec: builtins.int = ...,
            frame_id: builtins.bytes = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["frame_id", b"frame_id", "seq", b"seq", "timestamp_nsec", b"timestamp_nsec", "timestamp_sec", b"timestamp_sec"]) -> None: ...

    @typing.final
    class SickScanPointFieldMsg(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NAME_FIELD_NUMBER: builtins.int
        OFFSET_FIELD_NUMBER: builtins.int
        DATATYPE_FIELD_NUMBER: builtins.int
        COUNT_FIELD_NUMBER: builtins.int
        name: builtins.bytes
        offset: builtins.int
        datatype: builtins.int
        count: builtins.int
        def __init__(
            self,
            *,
            name: builtins.bytes = ...,
            offset: builtins.int = ...,
            datatype: builtins.int = ...,
            count: builtins.int = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["count", b"count", "datatype", b"datatype", "name", b"name", "offset", b"offset"]) -> None: ...

    @typing.final
    class SickScanUint8Array(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        CAPACITY_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        BUFFER_FIELD_NUMBER: builtins.int
        capacity: builtins.int
        size: builtins.int
        buffer: builtins.bytes
        def __init__(
            self,
            *,
            capacity: builtins.int = ...,
            size: builtins.int = ...,
            buffer: builtins.bytes = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["buffer", b"buffer", "capacity", b"capacity", "size", b"size"]) -> None: ...

    @typing.final
    class SickScanPointFieldArray(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        CAPACITY_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        BUFFER_FIELD_NUMBER: builtins.int
        capacity: builtins.int
        size: builtins.int
        @property
        def buffer(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SickScanPointCloudMsg.SickScanPointFieldMsg]: ...
        def __init__(
            self,
            *,
            capacity: builtins.int = ...,
            size: builtins.int = ...,
            buffer: collections.abc.Iterable[global___SickScanPointCloudMsg.SickScanPointFieldMsg] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["buffer", b"buffer", "capacity", b"capacity", "size", b"size"]) -> None: ...

    HEADER_FIELD_NUMBER: builtins.int
    HEIGHT_FIELD_NUMBER: builtins.int
    WIDTH_FIELD_NUMBER: builtins.int
    FIELDS_FIELD_NUMBER: builtins.int
    IS_BIGENDIAN_FIELD_NUMBER: builtins.int
    POINT_STEP_FIELD_NUMBER: builtins.int
    ROW_STEP_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    IS_DENSE_FIELD_NUMBER: builtins.int
    NUM_ECHOS_FIELD_NUMBER: builtins.int
    SEGMENT_IDX_FIELD_NUMBER: builtins.int
    height: builtins.int
    width: builtins.int
    is_bigendian: builtins.int
    point_step: builtins.int
    row_step: builtins.int
    is_dense: builtins.int
    num_echos: builtins.int
    segment_idx: builtins.int
    @property
    def header(self) -> global___SickScanPointCloudMsg.SickScanHeader: ...
    @property
    def fields(self) -> global___SickScanPointCloudMsg.SickScanPointFieldArray: ...
    @property
    def data(self) -> global___SickScanPointCloudMsg.SickScanUint8Array: ...
    def __init__(
        self,
        *,
        header: global___SickScanPointCloudMsg.SickScanHeader | None = ...,
        height: builtins.int = ...,
        width: builtins.int = ...,
        fields: global___SickScanPointCloudMsg.SickScanPointFieldArray | None = ...,
        is_bigendian: builtins.int = ...,
        point_step: builtins.int = ...,
        row_step: builtins.int = ...,
        data: global___SickScanPointCloudMsg.SickScanUint8Array | None = ...,
        is_dense: builtins.int = ...,
        num_echos: builtins.int = ...,
        segment_idx: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["data", b"data", "fields", b"fields", "header", b"header"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["data", b"data", "fields", b"fields", "header", b"header", "height", b"height", "is_bigendian", b"is_bigendian", "is_dense", b"is_dense", "num_echos", b"num_echos", "point_step", b"point_step", "row_step", b"row_step", "segment_idx", b"segment_idx", "width", b"width"]) -> None: ...

global___SickScanPointCloudMsg = SickScanPointCloudMsg



"""This script decodes the binary files generated
by the Oak-D Camera
There are mainly three types of binary files generated by the Oak-D Camera:
1. Hyperspectral images
2. IMU data
3. Calibration data

"""

from __future__ import annotations

import csv
import struct
from pathlib import Path
from typing import Any

import cv2
from turbojpeg import TurboJPEG

from phenomate_core.preprocessing.base import BasePreprocessor
from phenomate_core.preprocessing.oak_d import oak_pb2

# Initialize TurboJPEG
image_decoder = TurboJPEG()  # type: ignore[no-untyped-call]


class OakFramePreprocessor(BasePreprocessor[bytes]):
    def __init__(self, path: str | Path, in_ext: str = "bin") -> None:
        super().__init__(path, in_ext)
        self.metadata: list[oak_pb2.OakImageMeta] = []

    def extract(self, **kwargs: Any) -> None:
        with self.path.open("rb") as f:
            while True:
                # Read the length of serialized meta
                # first 8 bytes of the file are
                # the length of the meta data
                serialized_timestamp = f.read(8)
                if not serialized_timestamp:
                    break
                system_timestamp = struct.unpack("d", serialized_timestamp)[0]
                self.system_timestamps.append(system_timestamp)

                length_bytes = f.read(4)
                if not length_bytes:
                    break
                length = int.from_bytes(length_bytes, byteorder="little")

                # Reading the serialized meta
                serialized_meta = f.read(length)
                meta = oak_pb2.OakImageMeta()
                meta.ParseFromString(serialized_meta)
                self.metadata.append(meta)

                # Read the length of the next serialized image data
                length_bytes = f.read(4)
                if not length_bytes:
                    break
                length = int.from_bytes(length_bytes, byteorder="little")

                # Read the serialized image data
                image_data = f.read(length)

                self.images.append(image_data)

    def save_image_metadata_to_csv(self, path: Path) -> None:
        """This function saves the image metadata to a CSV file."""
        # Define the CSV header
        header = [
            "amiga_system_timestamp",
            "instance_num",
            "sequence_num",
            "timestamp",
            "timestamp_device",
            "exposure_time",
            "iso_value",
            "lens_pos",
            "profile",
            "cbr_preferred_bitrate_kbps",
            "vbr_or_mjpeg_quality",
            "frames_per_keyframe",
            "RateControlMode",
        ]
        file_path = path / self.get_output_name(None, "csv")
        # Open the CSV file for writing
        with file_path.open(mode="w", newline="") as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=header)

            # Write the header
            writer.writeheader()

            # Write the metadata
            for meta, timestamp in zip(self.metadata, self.system_timestamps, strict=False):
                row = {
                    "amiga_system_timestamp": timestamp,
                    "instance_num": meta.instance_num,
                    "sequence_num": meta.sequence_num,
                    "timestamp": meta.timestamp,
                    "timestamp_device": meta.timestamp_device,
                    "exposure_time": meta.settings.exposure_time,
                    "iso_value": meta.settings.iso_value,
                    "lens_pos": meta.settings.lens_pos,
                    "profile": meta.encoder_options.profile,
                    "cbr_preferred_bitrate_kbps": meta.encoder_options.cbr_preferred_bitrate_kbps,
                    "vbr_or_mjpeg_quality": meta.encoder_options.vbr_or_mjpeg_quality,
                    "frames_per_keyframe": meta.encoder_options.frames_per_keyframe,
                    "RateControlMode": meta.encoder_options.rate_control_mode,
                }
                writer.writerow(row)

    def save_image_as_jpeg(self, path: Path) -> None:
        """This function decodes the image data and saves it as a JPEG file."""
        for index, (image_data, meta_data, amiga_timestamp) in enumerate(
            zip(self.images, self.metadata, self.system_timestamps, strict=False)
        ):
            # Decode the image data
            img = image_decoder.decode(image_data)  # type: ignore[no-untyped-call]
            # Extract metadata
            instance_num = meta_data.instance_num
            sequence_num = meta_data.sequence_num
            timestamp = meta_data.timestamp
            timestamp_device = meta_data.timestamp_device
            exposure_time = meta_data.settings.exposure_time
            iso_value = meta_data.settings.iso_value
            lens_pos = meta_data.settings.lens_pos

            # Prepare metadata text
            metadata_text = (
                f"Amiga Timestamp: {amiga_timestamp}\n"
                f"Instance Num: {instance_num}\n"
                f"Sequence Num: {sequence_num}\n"
                f"Timestamp: {timestamp}\n"
                f"Timestamp Device: {timestamp_device}\n"
                f"Exposure Time: {exposure_time}\n"
                f"ISO Value: {iso_value}\n"
                f"Lens Pos: {lens_pos}\n"
            )

            # Define text properties
            font = cv2.FONT_HERSHEY_SIMPLEX
            font_scale = 0.5
            font_color = (0, 0, 255)
            thickness = 1
            line_type = cv2.LINE_AA
            y0, dy = 30, 20

            # Writing metadata text to image
            for i, line in enumerate(metadata_text.split("\n")):
                y = y0 + i * dy
                cv2.putText(img, line, (10, y), font, font_scale, font_color, thickness, line_type)

            # Save the image as a JPEG file
            output_path = path / self.get_output_name(index, "jpeg")
            cv2.imwrite(str(output_path), img)

    def save(self, path: Path | str, **kwargs: Any) -> None:
        file_path = Path(path)
        file_path.mkdir(parents=True, exist_ok=True)
        self.save_image_metadata_to_csv(file_path)
        self.save_image_as_jpeg(file_path)


class OakImuPacketsPreprocessor(BasePreprocessor[oak_pb2.OakImuPacket]):
    def extract(self, **kwargs: Any) -> None:
        with self.path.open("rb") as f:
            while True:
                serialized_timestamp = f.read(8)
                if not serialized_timestamp:
                    break
                system_timestamp = struct.unpack("d", serialized_timestamp)[0]
                self.system_timestamps.append(system_timestamp)

                # Read the length of the next serialized message
                length_bytes = f.read(4)
                if not length_bytes:
                    break
                length = int.from_bytes(length_bytes, byteorder="little")

                serialized_data = f.read(length)
                if not serialized_data:
                    break

                # Deserialize the message
                imu_msg = oak_pb2.OakImuPackets()
                imu_msg.ParseFromString(serialized_data)

                # Extract the OakGyro and OakAccelero data
                for packet in imu_msg.packets:
                    self.images.append(packet)

    def save(self, path: Path | str, **kwargs: Any) -> None:
        file_path = Path(path) / self.get_output_name(None, "csv", "imu")
        with file_path.open(mode="w", newline="") as file:
            writer = csv.writer(file)

            # Write the header
            writer.writerow(
                [
                    "gyro_x",
                    "gyro_y",
                    "gyro_z",
                    "gyro_sequence_num",
                    "gyro_accuracy",
                    "gyro_timestamp",
                    "gyro_timestamp_device",
                    "gyro_timestamp_recv",
                    "accelero_x",
                    "accelero_y",
                    "accelero_z",
                    "accelero_sequence_num",
                    "accelero_accuracy",
                    "accelero_timestamp",
                    "accelero_timestamp_device",
                    "accelero_timestamp_recv",
                    "system_timestamp",
                ]
            )
            # Write the data
            for image, system_timestamp in zip(self.images, self.system_timestamps, strict=False):
                writer.writerow(
                    [
                        image.gyro_packet.gyro.x,
                        image.gyro_packet.gyro.y,
                        image.gyro_packet.gyro.z,
                        image.gyro_packet.sequence_num,
                        image.gyro_packet.accuracy,
                        image.gyro_packet.timestamp,
                        image.gyro_packet.timestamp_device,
                        image.gyro_packet.timestamp_recv,
                        image.accelero_packet.accelero.x,
                        image.accelero_packet.accelero.y,
                        image.accelero_packet.accelero.z,
                        image.accelero_packet.sequence_num,
                        image.accelero_packet.accuracy,
                        image.accelero_packet.timestamp,
                        image.accelero_packet.timestamp_device,
                        image.accelero_packet.timestamp_recv,
                        system_timestamp,
                    ]
                )


class OakCalibrationPreprocessor(BasePreprocessor[oak_pb2.OakCalibration]):
    def extract(self, **kwargs: Any) -> None:
        with self.path.open("rb") as f:
            while True:
                # Read the length of the next serialized message
                length_bytes = f.read(4)
                if not length_bytes:
                    break
                length = int.from_bytes(length_bytes, byteorder="little")

                # Read the serialized message
                serialized_data = f.read(length)
                if not serialized_data:
                    break

                # Deserialize the message
                cal_msg = oak_pb2.OakCalibration()
                cal_msg.ParseFromString(serialized_data)
                self.images.append(cal_msg)

    def save(self, path: str | Path, **kwargs: Any) -> None:
        file_path = Path(path) / self.get_output_name(None, "txt", "calibration")
        with file_path.open(mode="a") as file:
            for image in enumerate(self.images):
                file.write(str(image) + "\n")
